/* 
Bill Lazure, W2EB 
This program is designed to run a SA-612-based transceiver for use
in portable/rugged applications, based on KD1JV's ATS3-B and MTR 
design.  I chose the Arduino platform because it uses the SI5351 which is 
more versatile than the Analog Devices AD985X series of DDS and uses 
considerably less current than the high-speed crystal oscillators needed
to run the AD devices.
     This program is a heavily modified version of VK3HN's SP-6 program.
*/

//libraries
#include <Rotary.h>
#include <si5351.h>     // Etherkit si3531 library from NT7S,  V2.1.4   https://github.com/etherkit/Si5351Arduino 
#include <Wire.h>
#include <EEPROM.h>
#include "SSD1306Ascii.h"  //I Changed SSD1306Init.h to lower contrast.  If that causes a problem, switch back
#include "SSD1306AsciiAvrI2c.h"

// ------------------------------------------------------------------------------------------------------------------
// Pin assignments

#define ENCODER_B          3  // Encoder pin B
#define ENCODER_A          2  // Encoder pin A
#define PIN_TONE_OUT       5  // digital pin with keyed audio tone on it
#define SWR_BUTTON         8  // Low = SWR Function
#define ENCODER_BUTTON     9  // the digital input used to read the encoder's pushbutton
#define RECEIVE_ENABLE     10 // controls the T/R Switch (high = receive)
#define PA_ENABLE          13 // this pin allows current to flow to the P.A. mosfets
#define KEY                11 // input to the Arduino from the KEY (low == transmit)
#define RX_MUTE            12 // receiver mute 

//Analog pins
#define BAND_INPUT         A2 //analog input to tell what band the filter plugin is for
#define BATTERY_PIN        A6 // Analog pin for Input DC sensing circuit
int pwr_val = 0;              // stores the last reading of the Input DC Voltage sensing input
int fwd_val = 0;			  // stores the last reading of the FWD RF power sensing input 
#define PIN_SWR_FWD        A7 // Analog pin for SWR bridge forward
#define PIN_SWR_REV        A3 // Analog pin for SWR bridge reverse


volatile uint32_t bfo = 4194105ULL;  // the reference BFO frequency
#define I2C_OLED_ADDRESS 0x3C 
#define RST_PIN    -1                // Define proper RST_PIN if required
  SSD1306AsciiAvrI2c oled;
#define NBR_VFOS         5           // number of selectable Bands 
#define BUTTON_HELD_MS 700           // button down period in mS required to activate 2nd pushbutton function (not used)

// ------------------------------------------------------------------------------------------------------------------

#define CW_TONE_HZ       600  // CW tone frequency (Hz)

// struct for 'Band parameter set' records -- the parameters that will change with each Band
typedef struct {
  boolean  active;
  uint32_t vfo;
  uint32_t radix;
} VFOset_type;

VFOset_type VFOSet[NBR_VFOS]; // array of band parameter sets
byte v;                       // index into VFOSet array (representing the current VFO)
byte v_prev;


Si5351 si5351;                // I2C address defaults to x60 in the NT7S lib
Rotary r = Rotary(ENCODER_A, ENCODER_B);

//------------------------------------------------------------------------------------------------------------------

// variables for transmit-receive control 
bool mode_tx = false; 

// button variables used in main loop for sensing the multiplexed buttons
byte button_nbr; 
byte old_button_nbr = 0; 


bool func_button_pressed = false; // if true, the next button pressed is interpreted as a Function 
bool BFO_tune_flg = false;       // BFO Tune feature

// ------------------------------------------------------------------------------------------------------------------
// variables for controlling EEPROM writes
unsigned long last_freq_change_ms;
bool eeprom_written_since_last_freq_change; 
bool changed_f = false;


/**************************************/
/* Interrupt service routine for encoder frequency change */
/**************************************/
ISR(PCINT2_vect) {
  unsigned char result = r.process();
  if (result == DIR_CW)
    set_frequency(1);
  else if (result == DIR_CCW)
    set_frequency(-1);
}

/**************************************/
/* Change frequency; dir=1 Increment; dir=-1 Decrement
/**************************************/
void set_frequency(short dir)
{
  if(mode_tx) return;  // dial locks in transmit
  if (dir == 1)
  {
      VFOSet[v].vfo += VFOSet[v].radix;         
  }
  else 
  {
     if (dir == -1)
    VFOSet[v].vfo -= VFOSet[v].radix;        
  };
  changed_f = true;  
};


int read_Analog_pin(byte p)
{
// Take an averaged reading of Analog pin 'p'  
  int i, val=0, nbr_reads=2; 
  for (i=0; i<nbr_reads; i++)
  {
    val += analogRead(p);
  }
  return val/nbr_reads; 
};

// ------------------------------------------------------------------------------------------------------------------
// display routines

void update_freq()
{
  oled.setCursor(0, 0);
  unsigned long f_Hz = VFOSet[v].vfo;
  uint16_t f = f_Hz/1000;                    // frequency in kHz
  int mhz = f/1000;                          // MHz (0..99)
  int khz = f - mhz*1000;                    // kHz (0..999)
  int hz = f_Hz - (mhz*1000000 + khz*1000);  // Hz (0..999)
  
  if(mhz<10) oled.print(" ");
  oled.print(mhz, DEC);
  oled.print(",");
  if(khz<10) oled.print("0");
  if(khz<100) oled.print("0");
  oled.println(khz, DEC);
  oled.setCursor(66, 0);  
  oled.print(","); 
  if(hz<10) oled.print("0"); 
  if(hz<100) oled.print("0"); 
  oled.println(hz, DEC); 
}

void refresh_OLED(){   // Update the OLED display 

  oled.set1X();  

  update_freq();
  
  oled.set1X();
  oled.setCursor(0, 2);

    if(VFOSet[v].radix==10000) oled.print("10kHz     ");
    if(VFOSet[v].radix==1000)  oled.print("1kHz      ");
    if(VFOSet[v].radix==100)   oled.print("100Hz     ");
    if(VFOSet[v].radix==10)    oled.print("10Hz      ");

    oled.setCursor(0, 4);
    oled.print("              ");
 }

void battery_mtr()   //Read Input DC, convert and display
{   
    pwr_val = read_Analog_pin((byte)BATTERY_PIN); //read input DC
   
   float var = ((pwr_val*.01466)+.3); // Perform math on that pin voltage to convert to actual volts and put in variable "var"
   oled.setCursor(0, 6);  //Setup the display location
   oled.print("BAT");
   oled.setFont(fixed_bold10x15);
   oled.setCursor(35, 6); //position for the battery value
   oled.print(var, 2);    //display the contents of "var" to 2 digits
   oled.print("    ");    // display spaces to clear previous readings that may have been larger
 
};

void update_eeprom()
{
  if(abs( millis() - last_freq_change_ms ) > 5000)
  {
    if(eeprom_written_since_last_freq_change == false)
    {
      // do the eeprom write
      EEPROM.update(0, v);   // write the band index (v) to the first byte
        
      int element_len = sizeof(VFOset_type);
      for(int i=0; i<NBR_VFOS ; i++)    // write each element of the VFOSet array
      {
        EEPROM.put(1 + (i * element_len), VFOSet[i]);
      }
      eeprom_written_since_last_freq_change = true;
    }
  }
};


// ------------------------------------------------------------------------------------------------------------------
// TR switching and sequencing 

void receive_to_TRANSMIT()
{
   mode_tx = true;
   digitalWrite(RX_MUTE, 0);     // RX_Mute function is reversed.  Low mutes the receiver and enables sidetone
   tone(PIN_TONE_OUT, CW_TONE_HZ);
 
   si5351.output_enable(SI5351_CLK1, 0);  // turn the receive VFO off
   volatile uint32_t f; //Declare the temporary variable "f"
   f = VFOSet[v].vfo;   //place our transmit frequency into "f"
          
   si5351.set_freq(f * SI5351_FREQ_MULT, SI5351_CLK2); //convert "f" for the si5351, clock 2
   si5351.output_enable(SI5351_CLK2, 1); // turn the transmit VFO "ON"

   digitalWrite(RECEIVE_ENABLE, 0); // switch T/R
   digitalWrite(PA_ENABLE, 1);  // Power "ON" the PA
   delay(10); //wait very briefly
	
	//Read FWD RF power, convert and display
  fwd_val = read_Analog_pin((byte)PIN_SWR_FWD);
	float val = ((fwd_val*.03315)*(fwd_val*.03315)/100); //Formula determined empirically; VERY approximate
  oled.setCursor(0, 4); //place the cursor at the beginning of the 4th line
  oled.print("PWR ");  //write the FWD Power the OLED
  oled.setCursor(40, 4); //place the cursor past the word FWD
  oled.print(val, 2); //write my value to the second decimal
  oled.print("W   ");  //place the "W" (watts) after the numbers, then add spaces to clear any old info
  }

void TRANSMIT_to_receive()
{
  mode_tx = false;
  digitalWrite(PA_ENABLE, 0);  // Disable the PA
  digitalWrite(RECEIVE_ENABLE, 1); // Switch T/R to receive
  noTone(PIN_TONE_OUT); //Kill sidetone
    si5351.output_enable(SI5351_CLK2, 0); //Kill TX VFO
    si5351.output_enable(SI5351_CLK1, 1);  //  turn the Rx VFO back on again 
    digitalWrite(RX_MUTE, 1);     // RX_MUTE function is reversed.  High un-mutes the receiver
	//Clear the FWD Power display
    oled.setCursor(0, 4); //place the cursor at the beginning of the 4th line
    oled.print("PWR 0.00W    ");  //write 0.00 FWD Power

};

void READ_SWR()
{
//-- VSWR meter code -------------------------------------------
    int proc = 1; //temporary variable used to contain status of "SWR" button
    int XYZ = 0; //temporary variable used to tell if SWR routine has run
    float fwd_max = 0; //initialize all computation variables
    float rev_max = 0;
    float n; 
    float SWR = 0.00;
 
  digitalWrite(RX_MUTE, 0);     // Mute the receiver. RX_MUTE function is reversed: Low mutes the receiver and enables sidetone.
  tone(PIN_TONE_OUT, CW_TONE_HZ); //Send the sidetone
      si5351.output_enable(SI5351_CLK1, 0);  // turn the receive VFO off
    volatile uint32_t f; 
    f = VFOSet[v].vfo; 
    
    si5351.set_freq(f * SI5351_FREQ_MULT, SI5351_CLK2); //set the tx frequency
    si5351.output_enable(SI5351_CLK2, 1); // turn the transmit VFO "ON"

    digitalWrite(RECEIVE_ENABLE, 0); // Disconnect the Receiver
    digitalWrite(PA_ENABLE, 1);  // Power "ON" the PA
    delay(10); //Let the transmitter settle down a bit before taking readings.
    
  proc = digitalRead(SWR_BUTTON); //put SWR button status in Proc
  while(proc == 0) //as long as the SWR button is pressed....
{
  delay(10);
  n = read_Analog_pin((byte)PIN_SWR_FWD);
       fwd_max = n;     

  n = read_Analog_pin((byte)PIN_SWR_REV)+25;
       rev_max = n;

    SWR = (fwd_max + rev_max)/(fwd_max-rev_max); //the actual SWR calculation

  oled.setCursor(0, 4); //write the SWR value to the OLED
  oled.print("SWR  ");
  oled.setCursor(40, 4);
  oled.print(SWR, 2);
  oled.print(":1   ");

    proc = digitalRead(SWR_BUTTON);  //read the SWR button again to see if it has changed
    XYZ = 1; //tell the rest of the program that the SWR routine ran
};
    
   if(XYZ == 1)  //Did the SWR routine run?
  {
 oled.setCursor(0, 4);
  oled.print("               "); //Blank the SWR lines from the OLED

    digitalWrite(PA_ENABLE, 0);  // Disable the PA
    digitalWrite(RECEIVE_ENABLE, 1); // Switch T/R to receive
    noTone(PIN_TONE_OUT); //Kill sidetone
    si5351.output_enable(SI5351_CLK2, 0); //Kill TX VFO
    si5351.output_enable(SI5351_CLK1, 1);  //  turn the Rx VFO back on again 
    digitalWrite(RX_MUTE, 1);     // RX_MUTE function is reversed.  High un-mutes the receiver
}; //resume normal program operation.

}


void setup()
{
  Serial.begin(9600);  
  Wire.begin();

  oled.begin(&Adafruit128x64, I2C_OLED_ADDRESS);
  oled.setFont(fixed_bold10x15);
  oled.clear();
  oled.println("    W2EB     ");
  oled.println(" GoLiaTH ");
  oled.println("   Final   ");
  oled.println("V20240328  ");

  delay(1000);
  oled.clear();

  pinMode(ENCODER_A, INPUT_PULLUP); // Encoder Pullup
  pinMode(ENCODER_B, INPUT_PULLUP); // Encoder Pullup
  pinMode(ENCODER_BUTTON, INPUT_PULLUP); //Pullup for the encoder switch
  pinMode(PA_ENABLE, OUTPUT);  // this pin enables current flow to the PA transistors
  pinMode(KEY, INPUT_PULLUP);  // enable pullup resistor for the KEY input
  pinMode(RX_MUTE, OUTPUT); 
  pinMode(SWR_BUTTON, INPUT_PULLUP); //enable pullup resistor for the SWR switch

  mode_tx = false;
  
  PCICR |= (1 << PCIE2);           // Enable pin change interrupt for the encoder
  PCMSK2 |= (1 << PCINT18) | (1 << PCINT19);
  sei();
  

  // load up VFOSet array from EEPROM

  v = EEPROM.read(0);
    Serial.print("setup() eeprom: v=");
  Serial.println(v);
  if(v >= NBR_VFOS) v = 1;  // in case NBR_VFOS has been reduced since the last run (EEPROM update)
  v_prev = v;
  
  int element_len = sizeof(VFOset_type);
  for(int i=0; i < NBR_VFOS; i++)
  {
    EEPROM.get(1 + (i * element_len), VFOSet[i]);
  };
 
// initialise and start the si5351 clocks

  si5351.init(SI5351_CRYSTAL_LOAD_8PF, 0, 130000); // 130000 is the correction factor for *my* hardware.  Yours may vary                         
  si5351.set_pll(SI5351_PLL_FIXED, SI5351_PLLA);
  si5351.drive_strength(SI5351_CLK2, SI5351_DRIVE_2MA); // set Tx VFO drive strength
  si5351.drive_strength(SI5351_CLK1, SI5351_DRIVE_2MA); // set rx VFO drive strength
  si5351.output_enable(SI5351_CLK0, 0); //Kill unused BFO

  changed_f = true; 
  last_freq_change_ms = millis(); 
  eeprom_written_since_last_freq_change = false; 

  {
  byte b=0; 
  int z;
  z = read_Analog_pin((byte)BAND_INPUT); //below is the voltage ladder used to determine the band info
  if(z >= 1  && z < 204) v = 0;  // v = 0 is the 80M Band  
    if(z >= 205  && z < 409) v = 1;  // v = 1 is the 40M Band
      if(z >= 410  && z < 614) v = 2;  // v = 2 is the 20M Band 
        if(z >= 615  && z < 819) v = 3;  // v = 3 is the 15M Band 
  if(z > 820 && z <= 1024) v = 4;  // v = 4 is the 10M Band
  
     refresh_OLED();
}

  //setup the rig for "Receive" mode
  digitalWrite(PA_ENABLE, 0);  // Disable the PA
  digitalWrite(RECEIVE_ENABLE, 1); // Switch T/R to receive
  digitalWrite(RX_MUTE, 1);     // RX_MUTE function is reversed.  High un-mutes the receiver

}  // End Setup Block

// ------------------------------------------------------------------------------------------------------------------
// loop()

void loop()
{
  update_eeprom(); //Update only happens if something changed
   battery_mtr(); //Take a reading and display it
   
//--Establish our Local Oscillator frequency by adding the BFO to the stored Transmit Frequency

  if (!mode_tx and changed_f) //if we aren't transmitting and the freq has changed...
  {
    volatile uint32_t f; 
     f = VFOSet[v].vfo; 
      si5351.set_freq((bfo + f) * SI5351_FREQ_MULT, SI5351_CLK1); //Add BFO to freq for receive...makes all sigs USB
      update_freq();

       
    changed_f = false; //reset the "frequency has changed" flag
    last_freq_change_ms = millis(); 
    eeprom_written_since_last_freq_change = false; //reset the "It's time to update the EEPROM" flag
  } 

  // sense the Key and switch the transmitter based on it 
      if(!digitalRead(KEY) && !mode_tx) 
    {
      receive_to_TRANSMIT(); // Key pressed
     };

    if(digitalRead(KEY)  &&  mode_tx) 
    {
      TRANSMIT_to_receive(); // Key released
       }

    if(digitalRead(SWR_BUTTON)==0)
    {
      READ_SWR();
    }
  
//--"Radix" Code------------------------------------------------------   
 if(!digitalRead(9)){         // The rotary encoder's pushbutton is used to change radix
    while(!digitalRead(9)) ;   // stay here until encoder button is released
   {        // default radfix increment/decrement behaviour...
      switch (VFOSet[v].radix)
      {
        case 10:
        {
            if(!func_button_pressed)
            {                // change radix up
                VFOSet[v].radix = 10000;
                // clear residual < 1kHz frequency component from the active VFO         
                uint16_t f = VFOSet[v].vfo % 1000;
                VFOSet[v].vfo -= f;
                    refresh_OLED();
            }
            else
            {
               func_button_pressed = false;
               // change radix down
               VFOSet[v].radix = 100;
               // clear residual < 100Hz frequency component from the active VFO         
               uint16_t f = VFOSet[v].vfo % 100;
               VFOSet[v].vfo -= f;
            }
        }
        break;
  
        case 100:
        {
            if(!func_button_pressed)
            {
                VFOSet[v].radix = 10;
                    refresh_OLED();
            }
            else
            {
                func_button_pressed = false;
                VFOSet[v].radix = 1000;
                // clear residual < 1kHz frequency component from the active VFO         
                uint16_t f = VFOSet[v].vfo % 1000;
                VFOSet[v].vfo -= f;
            }
        }
        break;
        
        case 1000:
        {
            if(!func_button_pressed)
            {
                VFOSet[v].radix = 100;
                    refresh_OLED();
            }
            else
            {
               func_button_pressed = false;
               VFOSet[v].radix = 10000;            
            }
            break;
        }
      
        case 10000:
        {
            if(!func_button_pressed)
            {
                VFOSet[v].radix = 1000;
                    refresh_OLED();
            }
            else
            {
                func_button_pressed = false;
                VFOSet[v].radix = 10;
            }
            break;
        }
      }
    }
} 

  
} //End LOOP
